/*
 *	ultrastor.c	Copyright (C) 1992 David B. Gentzel
 *	Low-level SCSI driver for UltraStor 14F
 *	by David B. Gentzel, Whitfield Software Services, Carnegie, PA
 *	    (gentzel@nova.enet.dec.com)
 *  scatter/gather added by Scott Taylor (n217cg@tamuts.tamu.edu)
 *  24F support by John F. Carr (jfc@athena.mit.edu)
 *	Thanks to UltraStor for providing the necessary documentation
 */

/*
 * TODO:
 *	1. Cleanup error handling & reporting.
 *	2. Find out why scatter/gather is limited to 16 requests per command.
 *	3. Add multiple outstanding requests.
 *	4. See if we can make good use of having more than one command per lun.
 *	5. Test/improve/fix abort & reset functions.
 *	6. Look at command linking (mscp.command_link and
 *	   mscp.command_link_id).
 *	7. Fix 24F support to allow multiple adapters.
 */

/*
 * NOTES:
 *    The UltraStor 14F, 24F, and 34F are a family of intelligent, high
 *    performance SCSI-2 host adapters.  They all support command queueing
 *    and scatter/gather I/O.  Some of them can also emulate the standard
 *    WD1003 interface for use with OS's which don't support SCSI.  Here
 *    is the scoop on the various models:
 *	14F - ISA first-party DMA HA with floppy support and WD1003 emulation.
 *	14N - ISA HA with floppy support.  I think that this is a non-DMA
 *	      HA.  Nothing further known.
 *	24F - EISA Bus Master HA with floppy support and WD1003 emulation.
 *	34F - VL-Bus Bus Master HA with floppy support (no WD1003 emulation).
 *
 *    The 14F and 24F are supported by this driver.  An effort has been
 *    made to support the 34F.  It should work, but is untested.  The 24F
 *    support is still alpha code.
 *
 *    Places flagged with a triple question-mark are things which are either
 *    unfinished, questionable, or wrong.
 */

#include <linux/stddef.h>
#include <linux/string.h>
#include <linux/sched.h>
#include <linux/kernel.h>

#include <asm/io.h>
#include <asm/bitops.h>
#include <asm/system.h>
#include <asm/dma.h>

#define ULTRASTOR_PRIVATE	/* Get the private stuff from ultrastor.h */
#include "../blk.h"
#include "scsi.h"
#include "hosts.h"
#include "ultrastor.h"

#define FALSE 0
#define TRUE 1

static inline int find_and_clear_bit(unsigned short *field)
{
  int rv;
  asm("xorl %0,%0\n0:\tbsfw %1,%0\n\tjz 1f\n\tbtr %0,%1\n\tjc 2f\n\tjmp 0b\n1:\tint3\n2:"
      : "=&r" (rv) : "m" (*field));
  return rv;
}

#define outl(val, port)	\
  asm volatile("outl %1,%0" : : "d" ((unsigned short)port), "a" ((int)val))

#define inl(port)	\
  ({unsigned int _v;	\
    asm volatile("inl %1,%0" : "=a" (_v): "d" ((unsigned short)port));\
    _v;})


#define ULTRASTOR_DEBUG UD_ABORT

#define VERSION "1.3 alpha"

#define ARRAY_SIZE(arr) (sizeof (arr) / sizeof (arr)[0])

#define PACKED		__attribute__((packed))
#define ALIGNED(x)	__attribute__((aligned(x)))


/* The 14F uses an array of 4-byte ints for its scatter/gather list.
   These need not be aligned, but can be declared as "int" anyway in
   this structure because this does not force extra padding.  */

typedef struct {
	unsigned int address;
	unsigned int num_bytes;
} ultrastor_sg_list;


/* MailBox SCSI Command Packet.  Basic command structure for communicating
   with controller. */
struct mscp {
    unsigned char opcode: 3;		/* type of command */
    unsigned char xdir: 2;		/* data transfer direction */
    unsigned char dcn: 1;		/* disable disconnect */
    unsigned char ca: 1;		/* use cache (if available) */
    unsigned char sg: 1;		/* scatter/gather operation */
    unsigned char target_id: 3;		/* target SCSI id */
    unsigned char ch_no: 2;		/* SCSI channel (always 0 for 14f) */
    unsigned char lun: 3;		/* logical unit number */
    unsigned int transfer_data PACKED;	/* transfer data pointer */
    unsigned int transfer_data_length PACKED;	/* length in bytes */
    unsigned int command_link PACKED;	/* for linking command chains */
    unsigned char scsi_command_link_id;	/* identifies command in chain */
    unsigned char number_of_sg_list;	/* (if sg is set) 8 bytes per list */
    unsigned char length_of_sense_byte;
    unsigned char length_of_scsi_cdbs;	/* 6, 10, or 12 */
    unsigned char scsi_cdbs[12];	/* SCSI commands */
    unsigned char adapter_status;	/* non-zero indicates HA error */
    unsigned char target_status;	/* non-zero indicates target error */
    unsigned int sense_data PACKED;
    /* The following fields are for software, unlike the hardware fields
       above.  */
    void (*done)(Scsi_Cmnd *);
    Scsi_Cmnd *SCint;
    ultrastor_sg_list sglist[ULTRASTOR_14F_MAX_SG];
};


/* Port addresses (relative to the base address) */
#define U14F_PRODUCT_ID(port) ((port) + 0x4)
#define CONFIG(port) ((port) + 0x6)

/* Port addresses relative to the doorbell base address.  */
#define LCL_DOORBELL_MASK(port) ((port) + 0x0)
#define LCL_DOORBELL_INTR(port) ((port) + 0x1)
#define SYS_DOORBELL_MASK(port) ((port) + 0x2)
#define SYS_DOORBELL_INTR(port) ((port) + 0x3)


/* Used to store configuration info read from config i/o registers.  Most of
   this is not used yet, but might as well save it.

   This structure also holds port addresses that are not at the same offset
   on the 14F and 24F.

   This structure holds all data that must be duplicated to support multiple
   adapters.  */

static struct ultrastor_config
{
  unsigned short port_address;		/* base address of card */
  unsigned short doorbell_address;	/* base address of doorbell CSRs */
  unsigned short ogm_address;		/* base address of OGM */
  unsigned short icm_address;		/* base address of ICM */
  const void *bios_segment;
  unsigned char interrupt: 4;
  unsigned char dma_channel: 3;
  unsigned char bios_drive_number: 1;
  unsigned char heads;
  unsigned char sectors;
  unsigned char ha_scsi_id: 3;
  unsigned char subversion: 4;
  unsigned char revision;
  unsigned char slot;

  volatile int aborted;

  /* Our index in the host adapter array maintained by higher-level driver */
  int host_number;

  /* A pool of MSCP structures for this adapter, and a bitmask of
     busy structures.  (If ULTRASTOR_14F_MAX_CMDS == 1, a 1 byte
     busy flag is used instead.)  */

#if ULTRASTOR_14F_MAX_CMDS == 1
  unsigned char mscp_busy;
#else
  unsigned short mscp_free;
#endif
  struct mscp mscp[ULTRASTOR_14F_MAX_CMDS];
} config;


/* Allowed BIOS base addresses for 14f (NULL indicates reserved) */
static const void *const bios_segment_table[8] = {
    NULL,	     (void *)0xC4000, (void *)0xC8000, (void *)0xCC000,
    (void *)0xD0000, (void *)0xD4000, (void *)0xD8000, (void *)0xDC000,
};

/* Allowed IRQs for 14f */
static const unsigned char interrupt_table_14f[4] = { 15, 14, 11, 10 };

/* Allowed DMA channels for 14f (0 indicates reserved) */
static const unsigned char dma_channel_table_14f[4] = { 5, 6, 7, 0 };

/* Head/sector mappings allowed by 14f */
static const struct {
    unsigned char heads;
    unsigned char sectors;
} mapping_table[4] = { { 16, 63 }, { 64, 32 }, { 64, 63 }, { 64, 32 } };

#ifndef PORT_OVERRIDE
/* ??? A probe of address 0x310 screws up NE2000 cards */
static const unsigned short ultrastor_ports_14f[] = {
    0x330, 0x340, /*0x310,*/ 0x230, 0x240, 0x210, 0x130, 0x140,
};
#endif

static void ultrastor_interrupt(int cpl);
static inline void build_sg_list(struct mscp *, Scsi_Cmnd *SCpnt);

static int ultrastor_14f_detect(int hostnum)
{
    size_t i;
    unsigned char in_byte, version_byte = 0;
    struct config_1 {
      unsigned char bios_segment: 3;
      unsigned char removable_disks_as_fixed: 1;
      unsigned char interrupt: 2;
    unsigned char dma_channel: 2;
    } config_1;
    struct config_2 {
      unsigned char ha_scsi_id: 3;
      unsigned char mapping_mode: 2;
      unsigned char bios_drive_number: 1;
      unsigned char tfr_port: 2;
    } config_2;

#if (ULTRASTOR_DEBUG & UD_DETECT)
    printk("US14F: detect: called\n");
#endif

    /* If a 24F has already been configured, don't look for a 14F.  */
    if (config.bios_segment)
	return FALSE;

#ifndef PORT_OVERRIDE
    config.port_address = 0;
    for (i = 0; i < ARRAY_SIZE(ultrastor_ports_14f); i++) {
	config.port_address = ultrastor_ports_14f[i];
#endif

#if (ULTRASTOR_DEBUG & UD_DETECT)
	printk("US14F: detect: testing port address %03X\n", config.port_address);
#endif

	in_byte = inb(U14F_PRODUCT_ID(config.port_address));
	if (in_byte != US14F_PRODUCT_ID_0) {
#if (ULTRASTOR_DEBUG & UD_DETECT)
# ifdef PORT_OVERRIDE
	    printk("US14F: detect: wrong product ID 0 - %02X\n", in_byte);
# else
	    printk("US14F: detect: no adapter at port %03X\n", config.port_address);
# endif
#endif
#ifdef PORT_OVERRIDE
	    return FALSE;
#else
	    continue;
#endif
	}
	in_byte = inb(U14F_PRODUCT_ID(config.port_address) + 1);
	/* Only upper nibble is significant for Product ID 1 */
	if ((in_byte & 0xF0) != US14F_PRODUCT_ID_1) {
#if (ULTRASTOR_DEBUG & UD_DETECT)
# ifdef PORT_OVERRIDE
	    printk("US14F: detect: wrong product ID 1 - %02X\n", in_byte);
# else
	    printk("US14F: detect: no adapter at port %03X\n", config.port_address);
# endif
#endif
#ifdef PORT_OVERRIDE
	    return FALSE;
#else
	    continue;
#endif
	}
	version_byte = in_byte;
#ifndef PORT_OVERRIDE
	break;
    }
    if (i == ARRAY_SIZE(ultrastor_ports_14f)) {
# if (ULTRASTOR_DEBUG & UD_DETECT)
	printk("US14F: detect: no port address found!\n");
# endif
	return FALSE;
    }
#endif

#if (ULTRASTOR_DEBUG & UD_DETECT)
    printk("US14F: detect: adapter found at port address %03X\n",
	   config.port_address);
#endif

    /* All above tests passed, must be the right thing.  Get some useful
       info. */
    *(char *)&config_1 = inb(CONFIG(config.port_address + 0));
    *(char *)&config_2 = inb(CONFIG(config.port_address + 1));
    config.bios_segment = bios_segment_table[config_1.bios_segment];
    config.doorbell_address = config.port_address;
    config.ogm_address = config.port_address + 0x8;
    config.icm_address = config.port_address + 0xC;
    config.interrupt = interrupt_table_14f[config_1.interrupt];
    config.ha_scsi_id = config_2.ha_scsi_id;
    config.heads = mapping_table[config_2.mapping_mode].heads;
    config.sectors = mapping_table[config_2.mapping_mode].sectors;
    config.bios_drive_number = config_2.bios_drive_number;
    config.subversion = (version_byte & 0x0F);
    if (config.subversion == U34F)
	config.dma_channel = 0;
    else
	config.dma_channel = dma_channel_table_14f[config_1.dma_channel];

    if (!config.bios_segment) {
#if (ULTRASTOR_DEBUG & UD_DETECT)
	printk("US14F: detect: not detected.\n");
#endif
	return FALSE;
    }

    /* Final consistancy check, verify previous info. */
    if (config.subversion != U34F)
	if (!config.dma_channel || !(config_2.tfr_port & 0x2)) {
#if (ULTRASTOR_DEBUG & UD_DETECT)
	    printk("US14F: detect: consistancy check failed\n");
#endif
	    return FALSE;
	}

    /* If we were TRULY paranoid, we could issue a host adapter inquiry
       command here and verify the data returned.  But frankly, I'm
       exhausted! */

    /* Finally!  Now I'm satisfied... */
#if (ULTRASTOR_DEBUG & UD_DETECT)
    printk("US14F: detect: detect succeeded\n"
	   "  Port address: %03X\n"
	   "  BIOS segment: %05X\n"
	   "  Interrupt: %u\n"
	   "  DMA channel: %u\n"
	   "  H/A SCSI ID: %u\n"
	   "  Subversion: %u\n",
	   config.port_address, config.bios_segment, config.interrupt,
	   config.dma_channel, config.ha_scsi_id, config.subversion);
#endif
    config.host_number = hostnum;
    scsi_hosts[hostnum].this_id = config.ha_scsi_id;
    scsi_hosts[hostnum].unchecked_isa_dma = (config.subversion != U34F);

#if ULTRASTOR_14F_MAX_CMDS > 1
    config.mscp_free = ~0;
#endif

    if (request_irq(config.interrupt, ultrastor_interrupt)) {
	printk("Unable to allocate IRQ%u for UltraStor controller.\n",
	       config.interrupt);
	return FALSE;
    }
    if (config.dma_channel && request_dma(config.dma_channel)) {
	printk("Unable to allocate DMA channel %u for UltraStor controller.\n",
	       config.dma_channel);
	free_irq(config.interrupt);
	return FALSE;
    }
    scsi_hosts[hostnum].sg_tablesize = ULTRASTOR_14F_MAX_SG;
    printk("UltraStor: scatter/gather enabled.  Using %d SG lists.\n",
	   ULTRASTOR_14F_MAX_SG);

    return TRUE;
}

static int ultrastor_24f_detect(int hostnum)
{
  register int i;

#if (ULTRASTOR_DEBUG & UD_DETECT)
  printk("US24F: detect");
#endif

  /* probe each EISA slot at slot address C80 */
  for (i = 1; i < 15; i++)
    {
      unsigned char config_1, config_2;
      unsigned short addr = (i << 12) | ULTRASTOR_24F_PORT;

      if (inb(addr) != US24F_PRODUCT_ID_0 &&
	  inb(addr+1) != US24F_PRODUCT_ID_1 &&
	  inb(addr+2) != US24F_PRODUCT_ID_2)
	continue;

      config.revision = inb(addr+3);
      config.slot = i;
      if (! (inb(addr+4) & 1))
	{
#if (ULTRASTOR_DEBUG & UD_DETECT)
	  printk("U24F: found disabled card in slot %u\n", i);
#endif
	  continue;
	}
#if (ULTRASTOR_DEBUG & UD_DETECT)
      printk("U24F: found card in slot %u\n", i);
#endif
      config_1 = inb(addr + 5);
      config.bios_segment = bios_segment_table[config_1 & 7];
      switch(config_1 >> 4)
	{
	case 1:
	  config.interrupt = 15;
	  break;
	case 2:
	  config.interrupt = 14;
	  break;
	case 4:
	  config.interrupt = 11;
	  break;
	case 8:
	  config.interrupt = 10;
	  break;
	default:
	  printk("U24F: invalid IRQ\n");
	  return FALSE;
	}
      if (request_irq(config.interrupt, ultrastor_interrupt))
	{
	  printk("Unable to allocate IRQ%u for UltraStor controller.\n",
		 config.interrupt);
	  return FALSE;
	}
      /* BIOS addr set */
      /* base port set */
      config.port_address = addr;
      config.doorbell_address = addr + 12;
      config.ogm_address = addr + 0x17;
      config.icm_address = addr + 0x1C;
      config_2 = inb(addr + 7);
      config.ha_scsi_id = config_2 & 7;
      config.heads = mapping_table[(config_2 >> 3) & 3].heads;
      config.sectors = mapping_table[(config_2 >> 3) & 3].sectors;
#if (ULTRASTOR_DEBUG & UD_DETECT)
      printk("US24F: detect: detect succeeded\n"
	     "  Port address: %03X\n"
	     "  BIOS segment: %05X\n"
	     "  Interrupt: %u\n"
	     "  H/A SCSI ID: %u\n",
	     config.port_address, config.bios_segment,
	     config.interrupt, config.ha_scsi_id);
#endif
      config.host_number = hostnum;
      scsi_hosts[hostnum].this_id = config.ha_scsi_id;
      scsi_hosts[hostnum].unchecked_isa_dma = 0;
      scsi_hosts[hostnum].sg_tablesize = ULTRASTOR_14F_MAX_SG;

#if ULTRASTOR_14F_MAX_CMDS > 1
      config.mscp_free = ~0;
#endif

      /* Disable CSIR interrupts (not used).  */
      outb(0x82, LCL_DOORBELL_MASK(addr+12));
      outb(0x02, SYS_DOORBELL_MASK(addr+12));
      printk("UltraStor: scatter/gather enabled.  Using %d SG lists.\n",
	     ULTRASTOR_14F_MAX_SG);
      return TRUE;
    }
  return FALSE;
}

int ultrastor_detect(int hostnum)
{
  return ultrastor_14f_detect(hostnum) || ultrastor_24f_detect(hostnum);
}

const char *ultrastor_info(void)
{
    static char buf[64];

    if (config.slot)
      sprintf(buf, "UltraStor 24F SCSI @ Slot %u Port %03X BIOS %05X IRQ%u\n",
	      config.slot, config.port_address,
	      (int)config.bios_segment, config.interrupt);
    else if (config.subversion)
      sprintf(buf, "UltraStor 34F SCSI @ Port %03X BIOS %05X IRQ%u\n",
	      config.port_address, (int)config.bios_segment,
	      config.interrupt);

    else
      sprintf(buf, "UltraStor 14F SCSI @ Port %03X BIOS %05X IRQ%u DMA%u\n",
	      config.port_address, (int)config.bios_segment,
	      config.interrupt, config.dma_channel);
    return buf;
}

static inline void build_sg_list(register struct mscp *mscp, Scsi_Cmnd *SCpnt)
{
	struct scatterlist *sl;
	long transfer_length = 0;
	int i, max;

	sl = (struct scatterlist *) SCpnt->request_buffer;
	/* Save ourselves some casts; can eliminate when we don't have to
	   look at it anymore! */
	max = SCpnt->use_sg;
	for (i = 0; i < max; i++) {
		mscp->sglist[i].address = (unsigned int)sl[i].address;
		mscp->sglist[i].num_bytes = sl[i].length;
		transfer_length += sl[i].length;
	}
	mscp->number_of_sg_list = (char) SCpnt->use_sg;
	mscp->transfer_data = (unsigned int)mscp->sglist;
	/* ??? May not be necessary.  Docs are unclear as to whether transfer
	   length field is ignored or whether it should be set to the total
	   number of bytes of the transfer.  */
	mscp->transfer_data_length = transfer_length;
}

int ultrastor_queuecommand(Scsi_Cmnd *SCpnt, void (*done)(Scsi_Cmnd *))
{
    register struct mscp *mscp;
#if ULTRASTOR_14F_MAX_CMDS > 1
    int mscp_index;
#else
#define mscp_index 0
#endif
    unsigned char in_byte;

#if (ULTRASTOR_DEBUG & UD_COMMAND)
    printk("USx4F: queuecommand: called\n");
#endif

#if ULTRASTOR_14F_MAX_CMDS == 1
    /* mscp_busy should never be true here.  */
    if (tas(&config.mscp_busy))
	panic("UltraStor device driver called while busy");
#else
    mscp_index = find_and_clear_bit(&config.mscp_free);
#if ULTRASTOR_DEBUG & UD_MULTI_CMD
    printk("Ux4f: using MSCP %d, free = %x\n", mscp_index, config.mscp_free);
#endif
#endif

    mscp = &config.mscp[mscp_index];
    mscp->opcode = OP_SCSI;

    mscp->xdir = DTD_SCSI;
    mscp->dcn = FALSE;
    /* The Archive Viper 2525 doesn't work properly if cache is enabled.  */
    mscp->ca = scsi_device_type[config.host_number][SCpnt->target] != TYPE_TAPE;
    mscp->target_id = SCpnt->target;
    mscp->ch_no = 0;
    mscp->lun = SCpnt->lun;
    if (SCpnt->use_sg) {
	/* Set scatter/gather flag in SCSI command packet */
	mscp->sg = TRUE;
	build_sg_list(mscp, SCpnt);
    } else {
	/* Unset scatter/gather flag in SCSI command packet */
	mscp->sg = FALSE;
	mscp->transfer_data = (unsigned int)SCpnt->request_buffer;
	mscp->transfer_data_length = SCpnt->request_bufflen;
    }
    mscp->command_link = 0;		/*???*/
    mscp->scsi_command_link_id = 0;	/*???*/
    mscp->length_of_sense_byte = 0;	/*???*/
    mscp->length_of_scsi_cdbs = COMMAND_SIZE(*(unsigned char *)SCpnt->cmnd);
    memcpy(mscp->scsi_cdbs, SCpnt->cmnd, mscp->length_of_scsi_cdbs);
    mscp->adapter_status = 0;
    mscp->target_status = 0;
    mscp->sense_data = 0;			/*???*/

    /* Find free OGM slot (OGMINT bit is 0) */
    do
	in_byte = inb(LCL_DOORBELL_INTR(config.doorbell_address));
    while (!config.aborted && (in_byte & 1));
    if (config.aborted) {
#if (ULTRASTOR_DEBUG & (UD_COMMAND | UD_ABORT))
	printk("USx4F: queuecommand: aborted\n");
#endif
	/* ??? is this right? */
	return (config.aborted << 16);
    }
    SCpnt->host_scribble = (unsigned char *)mscp;

    /* Store pointer in OGM address bytes */

    outl(mscp, config.ogm_address);

    /* Issue OGM interrupt */
    if (config.slot) {
	/* Write OGM command register on 24F */
	outb(1, config.ogm_address - 1);
	outb(0x2, LCL_DOORBELL_INTR(config.doorbell_address));
    } else {
	outb(0x1, LCL_DOORBELL_INTR(config.doorbell_address));
    }

    mscp->done = done;
    mscp->SCint = SCpnt;

#if (ULTRASTOR_DEBUG & UD_COMMAND)
    printk("USx4F: queuecommand: returning\n");
#endif

    return 0;
}

int ultrastor_abort(Scsi_Cmnd *SCpnt, int code)
{
#if ULTRASTOR_14F_MAX_CMDS > 1
    unsigned int mscp_index;

#endif

#if ULTRASTOR_14F_MAX_CMDS == 1
    config.mscp_busy = FALSE;
#else
    mscp_index = ((struct mscp *)SCpnt->host_scribble) - config.mscp;
    if (mscp_index >= ULTRASTOR_14F_MAX_CMDS)
	panic("Ux4F aborting invalid MSCP");

    set_bit(mscp_index, &config.mscp_free);
#endif

#if (ULTRASTOR_DEBUG & UD_ABORT)
    printk("US14F: abort\n");
#endif

    config.aborted = (code ? code : DID_ABORT);


    return 0;
}

int ultrastor_reset(void)
{
#if 0
    unsigned char in_byte;
#endif

#if (ULTRASTOR_DEBUG & UD_RESET)
    printk("US14F: reset: called\n");
#endif

	/* ??? SCSI bus reset causes problems on some systems. */
#if 0
    /* Issue SCSI BUS reset */
    outb_p(0x20, LCL_DOORBELL_INTR(config.doorbell_address));

    /* Wait for completion... */
    do
	in_byte = inb_p(LCL_DOORBELL_INTR(config.doorbell_address));
    while (in_byte & 0x20);

    config.aborted = DID_RESET;
#endif

#if (ULTRASTOR_DEBUG & UD_RESET)
    printk("US14F: reset: returning\n");
#endif
    return 0;
}

int ultrastor_biosparam(int size, int dev, int *info)
{
    unsigned int s = config.heads * config.sectors;

    info[0] = config.heads;
    info[1] = config.sectors;
    info[2] = (size + (s - 1)) / s;
    if (info[2] > 1024)
	info[2] = 1024;
    return 0;
}

static void ultrastor_interrupt(int cpl)
{
    unsigned int status;
#if ULTRASTOR_14F_MAX_CMDS > 1
    unsigned int mscp_index;
#endif
    register struct mscp *mscp;
    void (*done)(Scsi_Cmnd *);
    Scsi_Cmnd *SCtmp;

    /* Clean ICM slot (set ICMINT bit to 0) */
    if (config.slot) {
#if (ULTRASTOR_DEBUG & UD_INTERRUPT)
	unsigned char icm_status = inb(config.icm_address - 1);
	/* If icm_status is nonzero there has been an error.
	   TODO: check icm_status.  */
	if (icm_status != 1)
	    printk("US24F: ICM status %x\n", icm_status);
#endif
	outb(0, config.icm_address - 1);
	outb(2, SYS_DOORBELL_INTR(config.doorbell_address));
    } else {
	outb(1, SYS_DOORBELL_INTR(config.doorbell_address));
    }
#if ULTRASTOR_14F_MAX_CMDS == 1
    mscp = &config.mscp[0];
#else
    mscp = (struct mscp *)inl(config.icm_address);
    mscp_index = mscp - config.mscp;
    if (mscp_index >= ULTRASTOR_14F_MAX_CMDS)
	panic("Ux24F interrupt: bad MSCP address");
#endif

    /* Save done locally and zero before calling.  This is needed as
       once we call done, we may get another command queued before this
       interrupt service routine can return. */
    done = mscp->done;
    mscp->done = 0;
    SCtmp = mscp->SCint;

    if (done == 0)
	panic("US14F: interrupt: unexpected interrupt");


#if (ULTRASTOR_DEBUG & UD_INTERRUPT)
    if (mscp->adapter_status)
	printk("U24F status %x:%x\n", config.mscp.adapter_status,
	       config.mscp.target_status);
#endif

    /* Let the higher levels know that we're done */
    switch (mscp->adapter_status)
      {
      case 0:
	status = DID_OK << 16;
	break;
      case 0x01:	/* invalid command */
      case 0x02:	/* invalid parameters */
      case 0x03:	/* invalid data list */
      default:
	status = DID_ERROR << 16;
	break;
      case 0x84:	/* SCSI bus abort */
	status = DID_ABORT << 16;
	break;
      case 0x91:
	status = DID_TIME_OUT << 16;
	break;
      }

    SCtmp->result = status | mscp->target_status;

    SCtmp->host_scribble = 0;

    /* Free up mscp block for next command */
#if ULTRASTOR_14F_MAX_CMDS == 1
    config.mscp_busy = FALSE;
#else
    set_bit(mscp_index, &config.mscp_free);
#endif
    
    done(SCtmp);

#if (ULTRASTOR_DEBUG & UD_INTERRUPT)
    printk("USx4F: interrupt: returning\n");
#endif
}
